small  ICON "nlarn-16.ico"
medium ICON "nlarn-32.ico"
large  ICON "nlarn-48.ico"

1 VERSIONINFO
FILEVERSION     VERSION_MAJOR,VERSION_MINOR,VERSION_PATCH,0
PRODUCTVERSION  VERSION_MAJOR,VERSION_MINOR,VERSION_PATCH,0
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "080904E4"
    BEGIN
      VALUE "CompanyName", "Joachim de Groot"
      VALUE "FileDescription", "NLarn"
      VALUE "FileVersion", VINFO
      VALUE "InternalName", "nlarn"
      VALUE "LegalCopyright", "(C) 2009-2020 Joachim de Groot"
      VALUE "OriginalFilename", "nlarn.exe"
      VALUE "ProductName", "NLarn"
      VALUE "ProductVersion", VINFO
    END
  END

  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x809, 1252
  END
END
